﻿using System.Data;
using System.Data.Common;
using System.Diagnostics.CodeAnalysis;
using System.Text;
using Connected.ServiceModel.Client.Data.Remote;

namespace Connected.ServiceModel.Client.Data;
internal sealed class TableDataCommand : DbCommand
{
	public TableDataCommand(RemoteTableService tables)
	{
		Parameters = new DataParameterCollection();
		Tables = tables;
	}

	public new DataParameterCollection Parameters { get; }
	public RemoteTableService Tables { get; }
	[AllowNull]
	public override string CommandText { get; set; }
	public override int CommandTimeout { get; set; } = 120;
	public override CommandType CommandType { get; set; } = CommandType.Text;
	protected override DbConnection? DbConnection { get; set; }
	protected override DbParameterCollection DbParameterCollection => Parameters;
	protected override DbTransaction? DbTransaction { get; set; }
	public override bool DesignTimeVisible { get; set; }
	public override UpdateRowSource UpdatedRowSource { get; set; }

	protected override DbParameter CreateDbParameter()
	{
		return new TableDataParameter();
	}

	public override async Task<int> ExecuteNonQueryAsync(CancellationToken cancellationToken)
	{
		await Tables.Update(ParseCommandText());

		return 0;
	}

	public override int ExecuteNonQuery()
	{
		AsyncUtils.RunSync(() => Tables.Update(ParseCommandText()));

		return 0;
	}

	protected override async Task<DbDataReader> ExecuteDbDataReaderAsync(CommandBehavior behavior, CancellationToken cancellationToken)
	{
		return new TableDataReader(await Tables.Query(ParseCommandText()));
	}

	protected override DbDataReader ExecuteDbDataReader(CommandBehavior behavior)
	{
		return new TableDataReader(AsyncUtils.RunSync(() => Tables.Query(ParseCommandText())));
	}

	private string ParseCommandText()
	{
		var text = new StringBuilder(CommandText);

		foreach (IDbDataParameter parameter in Parameters)
			text = text.Replace(parameter.ParameterName, ParseValue(parameter));

		return text.ToString();
	}

	private static string? ParseValue(IDbDataParameter parameter)
	{
		if (parameter.Value is null || parameter.Value == DBNull.Value)
			return "null";

		return parameter.DbType switch
		{
			DbType.AnsiString or DbType.String or DbType.AnsiStringFixedLength or DbType.StringFixedLength or DbType.Xml => $"'{parameter.Value}'",
			_ => parameter.Value.ToString(),
		};
	}

	public override void Cancel()
	{
	}

	public override object? ExecuteScalar()
	{
		var reader = ExecuteReader();

		if (reader.Read())
			return reader[0];

		return null;
	}

	public override void Prepare()
	{
	}
}
