﻿using System.Collections;
using System.Data.Common;
using System.Diagnostics.CodeAnalysis;
using System.Text.Json.Nodes;

namespace Connected.ServiceModel.Client.Data;
internal sealed class TableDataReader : DbDataReader
{
	public TableDataReader(JsonArray items)
	{
		Items = items;
	}

	public override object this[int ordinal] => throw new NotImplementedException();

	public override object this[string name] => throw new NotImplementedException();

	private JsonArray Items { get; }
	private int Index { get; set; } = -1;
	private JsonObject? Current => Index < 0 ? null : Items[Index] as JsonObject;
	public override int Depth => 1;
	public override int FieldCount => Current is null ? 0 : Current.Count;

	public override bool HasRows => Items.Any();

	public override bool IsClosed => false;

	public override int RecordsAffected => Items.Count;

	public override bool GetBoolean(int ordinal)
	{
		return GetValue<bool>(ordinal);
	}

	public override byte GetByte(int ordinal)
	{
		return GetValue<byte>(ordinal);
	}

	public override long GetBytes(int ordinal, long dataOffset, byte[]? buffer, int bufferOffset, int length)
	{
		var value = GetValue<string>(ordinal);
		var bytes = Convert.FromBase64String(value);

		buffer ??= new byte[bufferOffset + length];

		for (var i = 0; i < length; i++)
			buffer[bufferOffset + i] = bytes[dataOffset + i];

		return length;
	}

	public override char GetChar(int ordinal)
	{
		return GetValue<char>(ordinal);
	}

	public override long GetChars(int ordinal, long dataOffset, char[]? buffer, int bufferOffset, int length)
	{
		var value = GetValue<string>(ordinal);
		var bytes = Convert.FromBase64String(value);

		buffer ??= new char[bufferOffset + length];

		for (var i = 0; i < length; i++)
			buffer[bufferOffset + i] = (char)bytes[dataOffset + i];

		return length;
	}

	public override string GetDataTypeName(int ordinal)
	{
		/*
		 * TODO: resolve appropriate type
		 */
		return typeof(string).Name;
	}

	public override DateTime GetDateTime(int ordinal)
	{
		return GetValue<DateTime>(ordinal);
	}

	public override decimal GetDecimal(int ordinal)
	{
		return GetValue<decimal>(ordinal);
	}

	public override double GetDouble(int ordinal)
	{
		return GetValue<double>(ordinal);
	}

	public override IEnumerator GetEnumerator()
	{
		return Items.GetEnumerator();
	}

	[return: DynamicallyAccessedMembers(DynamicallyAccessedMemberTypes.PublicFields | DynamicallyAccessedMemberTypes.PublicProperties)]
	public override Type GetFieldType(int ordinal)
	{
		/*
		 * TODO: resolve appropriate type
		 */
		return typeof(string);
	}

	public override float GetFloat(int ordinal)
	{
		return GetValue<float>(ordinal);
	}

	public override Guid GetGuid(int ordinal)
	{
		return GetValue<Guid>(ordinal);
	}

	public override short GetInt16(int ordinal)
	{
		return GetValue<short>(ordinal);
	}

	public override int GetInt32(int ordinal)
	{
		return GetValue<int>(ordinal);
	}

	public override long GetInt64(int ordinal)
	{
		return GetValue<long>(ordinal);
	}

	public override string GetName(int ordinal)
	{
		if (Current is null)
			throw new NullReferenceException();

		if (Current.ElementAt(ordinal).Key is string key)
			return key;

		throw new NullReferenceException();
	}

	public override int GetOrdinal(string name)
	{
		if (Current is null)
			throw new NullReferenceException();

		for (var i = 0; i < Current.Count; i++)
		{
			if (string.Equals(Current.ElementAt(i).Key, name, StringComparison.OrdinalIgnoreCase))
				return i;
		}

		return -1;
	}

	public override string GetString(int ordinal)
	{
		return GetValue<string>(ordinal);
	}

	public override object GetValue(int ordinal)
	{
		return GetValue<object>(ordinal);
	}

	public override int GetValues(object[] values)
	{
		if (Current is null)
			throw new NullReferenceException();

		for (var i = 0; i < Current.Count; i++)
		{
			var value = Current.ElementAt(i);

			if (value.Value is not null)
				values[i] = value.Value.AsValue().GetValue<object>();
		}

		return Current.Count;
	}

	public override bool IsDBNull(int ordinal)
	{
		return false;
	}

	public override bool NextResult()
	{
		return false;
	}

	public override bool Read()
	{
		if (Items.Count >= Index + 1)
		{
			Index++;

			return true;
		}

		return false;
	}

	private T GetValue<T>(int ordinal)
	{
		return GetCurrentValue(ordinal).GetValue<T>();
	}

	private JsonValue GetCurrentValue(int ordinal)
	{
		if (Current is null)
			throw new NullReferenceException();

		if (Current.ElementAt(ordinal).Value is not JsonValue value)
			throw new NullReferenceException();

		return value;
	}
}
